function[Xp, Yp, infos] = dlmafit_mod(N, M, rk, L, Pomega, Pomega_test, W, options)
    
    % D-LMaFit options
    delta = 1;
    gam_min = 0.7;
    CUT = ones(1,L)*((M/L));
    
    % Adjency matrix
    Nivec = zeros(L^2,1);
    Nivec(sub2ind([L L],1 : L-1,2 : L)) = 1;
    Ni = reshape(Nivec, L, L);
    Ni = Ni + Ni';
    
    
    
    [P, s] = fdla_symmetric(Ni);      % mixing matrix P
    P_tilde = (eye(L)+P)/2;     % mixing matrix P_tilde
    
    
    % Sparsity pattern
    Pomega_i = mat2cell(Pomega,[N],[CUT]);
    Wi = mat2cell(W,[N],[CUT]);
    Pomega_test_i = mat2cell(Pomega_test,[N],[CUT]);
   
    % Initialization
    Xi = rand(N, rk, L);
    Yi = rand(rk, M/L, L);
    Z = Pomega.*W; Zi = mat2cell(Z, [N], [CUT]);
    
    
    Qp = zeros(N, M/L, L);
    X_oldold = zeros(N, rk, L);
    q_old = zeros(N, rk, L);
    q_oldold = zeros(N,rk,L);
    wi = ones(1, L);
    Be = zeros(N, M/L, L);
    
    
    for i = 1 : L
        numentries{i} = sum(sum(Pomega_i{i}));
        numentries_test{i} = sum(sum(Pomega_test_i{i}));
    end
    
    for ii = 1 : L
        infos{ii}.cost = [];
        infos{ii}.cost_test =[];
        infos{ii}.commcost=[];
    end
    
    if ~isfield(options,'iters'); options.iters = 100; end % 100 updates of all the agents.
    if ~isfield(options,'alpha'); options.alpha = 0.4; end % Stepsize.
    
    
    
    %% update
    for iter = 1 : options.iters
        Xp = Xi; Yp = Yi; Zp = Zi; wp = wi;
        
        for i = 1 : L
            Be(:,:,i) = Pomega_i{i}.*(Xp(:,:,i)*Yp(:,:,i)) - Pomega_i{i}.*Wi{i};    % residual before update
            Qp(:,:,i) = wp(i)*Zp{i}+(1-wp(i))*Xp(:,:,i)*Yp(:,:,i);                  % Qp stands for Z
            q_old(:,:,i) = Qp(:,:,i)*Yp(:,:,i)';
        end
        
        
        % update of Xp
        X_old = Xp;
        
        if iter == 1
            for i = 1:L
                sum_P = zeros(N,rk);
                
                for j = 1:L
                    sum_P = P(i,j)*X_old(:,:,j) + sum_P;
                end
                
                Xp(:,:,i)=sum_P-options.alpha*(X_old(:,:,i)-q_old(:,:,i));
            end
        end
        
        
        if iter >= 2
            for i = 1 : L
                sum_P = zeros(N,rk); sum_P_tilde=zeros(N,rk);
                for j = 1 : L
                    sum_P = P(i,j)*X_old(:,:,j) + sum_P;
                    sum_P_tilde=P_tilde(i,j)*X_oldold(:,:,i)+sum_P_tilde;
                end
                Xp(:,:,i)=X_old(:,:,i) + sum_P - sum_P_tilde - options.alpha*(X_old(:,:,i)-X_oldold(:,:,i)-q_old(:,:,i)+q_oldold(:,:,i));
            end
        end
        X_oldold =Xi; q_oldold = q_old;
        
        
        
        % update of Y,Z
        for i=1:L
            Yp(:,:,i) = (Xp(:,:,i)'*Xp(:,:,i))\(Xp(:,:,i)'*Qp(:,:,i));%  inv(Xp(:,:,i)'*Xp(:,:,i))*Xp(:,:,i)'*Qp(:,:,i);
            Zp{i} = Xp(:,:,i)*Yp(:,:,i) + Pomega_i{i}.*(Wi{i} - Xp(:,:,i)*Yp(:,:,i));
        end
        
        
        % update of wi
        XY = zeros(N, 1); %XY1=XY; XY2=XY;
        for i = 1 : L
            Af = Pomega_i{i}.*(Xp(:,:,i)*Yp(:,:,i))-Pomega_i{i}.*Wi{i};
            gami = norm(Af,'fro')/norm(Be(:,:,i),'fro');
            if (gami >= gam_min) && (gami<1)
                wi(i) = wi(i)+delta;
            elseif (gami >= 1)
                wi(i) = 1;
            end
            XYi = Xp(:,:,i)*Yp(:,:,i);
            XY = [XY, XYi];
        end
        XY(:,1)=[];
        Xi = Xp;
        Yi= Yp;
        Zi =Zp;
        wi = wp;
        
        
        for i = 1 : L
            infos{i}.cost =[infos{i}.cost; (1/numentries{i})*norm(Pomega_i{i}.*(Wi{i} - Xp(:,:,i)*Yp(:,:,i)),'fro')^2];
            infos{i}.cost_test =[infos{i}.cost_test; (1/numentries_test{i})*norm(Pomega_test_i{i}.*(Wi{i} - Xp(:,:,i)*Yp(:,:,i)),'fro')^2];

            if i ~= L                
                problem.M = grassmannfactory(N, rk);
                [Upi, ~] = qr(Xp(:, :, i), 0);
                [Upj, ~] = qr(Xp(:, :, j), 0);
                infos{i}.commcost = [infos{i}.commcost; problem.M.norm(Upi, problem.M.log(Upi, Upj))^2];
            end
        end
        
        cost1 = infos{1}.cost(end);
        cost2 = infos{2}.cost(end);
        commcost12 =  infos{1}.commcost(end);
        
        
        fprintf('Iteration %0.3d, cost node %d: %7.3e, cost node %d: %7.3e, communication cost: %7.3e\n', iter, 1, cost1, 2, cost2, commcost12);
        
    end
    
end