%% This code includes
%   centralized LMaFit(C-LMaFit) and decentralized LMaFit(D-LMaFit)
clear all
clc;
close all hidden;

%% random seed
%seed = round(5000*rand);               % use this line if you set a random seed
seed = 3302;                            % use this line if you set a fixed seed
fprintf('Seed = %d\n',seed);            % print the current seed 
if exist('RandStream','file')
   RandStream.setGlobalStream(RandStream('mt19937ar','seed',seed));
else
   rand('state',seed); 
   randn('state',seed^2);
end

%% Set parameters
L = 50;              % num of agents
N = 200; M = 2000;   % size of matrix   
K = 4; rk = K;       % rank of matrix
p = 0.4; sr = p;     % sampling ratio       
alpha = 0.4;         % stepsize   
cnt = 0.2;           % connect ratio of network
delta = 1; gam_min = 0.7; 
CUT = ones(1,L)*(M/L); 
Ni = Produce_Adjacency_Matrix(L, cnt);   % adjacency matrix Ni
[P,s] = fdla_symmetric(Ni);              % mixing matrix P
P_tilde = (eye(L)+P)/2;                  % mixing matrix P_tilde

%% Initialization
% D-LMaFit
Pomega = zeros(N,M); idx = randperm(N*M); Pomega(idx(1:N*M*sr)) = 1;    
Pomega_i = mat2cell(Pomega,[N],[CUT]);   
W = rand(N,rk)*diag(rk)*rand(rk,M);      
Wi = mat2cell(W,[N],[CUT]);              
Xi = rand(N,rk,L);                       
Yi = rand(rk,M/L,L);                     
Z = Pomega.*W; Zi = mat2cell(Z,[N],[CUT]);              
Qp = zeros(N,M/L,L);
X_old = zeros(N,rk,L);
X_oldold = zeros(N,rk,L);
q_old = zeros(N,rk,L);                
q_oldold = zeros(N,rk,L);                 
wi = ones(1,L);                        
Be = zeros(N,M/L,L);
% C-LMaFit
Be1 = Be;
Qp1 = Qp;
q_old1 = q_old;
Xi1 = Xi;                       
Yi1 = Yi; 
Zi1 = Zi;
wi1 = wi;

%% update
convergent = false; iter = 0; 
while ~convergent
    iter = iter+1;
    Xp = Xi; Yp = Yi; Zp = Zi; wp = wi; 
    Xp1 = Xi1; Yp1 = Yi1; Zp1 = Zi1; wp1 = wi1;                                    
    for i = 1:L
        Be(:,:,i) = Pomega_i{i}.*(Xp(:,:,i)*Yp(:,:,i))-Pomega_i{i}.*Wi{i};    % residual before update
        Qp(:,:,i) = wp(i)*Zp{i}+(1-wp(i))*Xp(:,:,i)*Yp(:,:,i);                % Qp stands for Z
        q_old(:,:,i) = Qp(:,:,i)*Yp(:,:,i)';          
        Be1(:,:,i) = Pomega_i{i}.*(Xp1(:,:,i)*Yp1(:,:,i))-Pomega_i{i}.*Wi{i}; 
        Qp1(:,:,i) = wp1(i)*Zp1{i}+(1-wp1(i))*Xp1(:,:,i)*Yp1(:,:,i);  
        q_old1(:,:,i) = Qp1(:,:,i)*Yp1(:,:,i)'; 
    end    
    % update of Xp
    X_old=Xp;
    if iter==1
        for i=1:L      
            sum_P=zeros(N,rk);
            for j=1:L           
                sum_P=P(i,j)*X_old(:,:,j)+sum_P;
            end
            Xp(:,:,i)=sum_P-alpha*(X_old(:,:,i)-q_old(:,:,i));
        end
    end    
    if iter>=2
        for i=1:L
            sum_P=zeros(N,rk); sum_P_tilde=zeros(N,rk);
            for j=1:L
                sum_P=P(i,j)*X_old(:,:,j)+sum_P;
                sum_P_tilde=P_tilde(i,j)*X_oldold(:,:,i)+sum_P_tilde;
            end
            Xp(:,:,i)=X_old(:,:,i)+sum_P-sum_P_tilde-alpha*(X_old(:,:,i)-X_oldold(:,:,i)-q_old(:,:,i)+q_oldold(:,:,i));
        end
    end      
    X_oldold=Xi; q_oldold=q_old;    
    % update of Xp1
    sum_q=zeros(N,rk);
    for i=1:L
        sum_q=sum_q+q_old1(:,:,i);
    end       
    for i=1:L
        Xp1(:,:,i)=1/L*sum_q;
    end     
    % update of Y,Z
    for i=1:L
        Yp(:,:,i)=inv(Xp(:,:,i)'*Xp(:,:,i))*Xp(:,:,i)'*Qp(:,:,i);
        Zp{i}=Xp(:,:,i)*Yp(:,:,i)+Pomega_i{i}.*(Wi{i}-Xp(:,:,i)*Yp(:,:,i));      
        Yp1(:,:,i)=inv(Xp1(:,:,i)'*Xp1(:,:,i))*Xp1(:,:,i)'*Qp1(:,:,i);
        Zp1{i}=Xp1(:,:,i)*Yp1(:,:,i)+Pomega_i{i}.*(Wi{i}-Xp1(:,:,i)*Yp1(:,:,i));      
    end       
    % update of wi
    XY=zeros(N,1); XY1=XY; XY2=XY;
    for i=1:L
        Af=Pomega_i{i}.*(Xp(:,:,i)*Yp(:,:,i))-Pomega_i{i}.*Wi{i};  
        gami=norm(Af,'fro')/norm(Be(:,:,i),'fro');
        if (gami>=gam_min)&&(gami<1)
            wi(i)=wi(i)+delta;
        elseif (gami>=1)
            wi(i)=1;
        end         
        XYi=Xp(:,:,i)*Yp(:,:,i);
        XY=[XY,XYi];
    end
    XY(:,1)=[];   
    Xi=Xp; Yi=Yp; Zi=Zp; wi=wp;    
    % update of wi1
    for i=1:L
    Af1=Pomega_i{i}.*(Xp1(:,:,i)*Yp1(:,:,i))-Pomega_i{i}.*Wi{i};  
    gami1=norm(Af1,'fro')/norm(Be1(:,:,i),'fro');
    if (gami1>=gam_min)&&(gami1<1)
        wi1(i)=wi1(i)+delta;
    elseif (gami1>=1)
        wi1(i)=1;
    end         
    XYi1=Xp1(:,:,i)*Yp1(:,:,i);
    XY1=[XY1,XYi1];
    end   
    XY1(:,1)=[];   
    Xi1=Xp1; Yi1=Yp1; Zi1=Zp1; wi1=wp1;   
    % normalized residual
    norres=norm((W-XY),'fro')/norm(W,'fro');        % D-LMaFit
    norres1=norm((W-XY1),'fro')/norm(W,'fro');      % C-LMaFit
    y_axis(iter)=norres;
    y_axis1(iter)=norres1;
    % stop criteria
    if (norres<10^(-20))||(iter>3000)
        convergent=true;
    end 
    fprintf('Iter: %d\t',iter);
    fprintf('D_nr= %8.3e\n ',norres);  
end

%% plot
% MakerNum = 24;
% x=zeros(1, MakerNum);
% y=zeros(1, MakerNum);
% y1=zeros(1, MakerNum);
% for i = 1:MakerNum
%     x(i) = 2^((i-1)*0.5);
%     y(i) = y_axis(fix(x(i)));
%     y1(i) = y_axis1(fix(x(i)));
% end
% loglog(x,y1,'--bo',...
%           'LineWidth',2,...
%           'MarkerEdgeColor','b',...
%           'MarkerFaceColor','b',...
%           'MarkerSize',8);
% hold on
% loglog(x,y,'--ks',...
%          'LineWidth',2,...
%          'MarkerEdgeColor','k',...
%          'MarkerFaceColor','k',...
%          'MarkerSize',8);
% hold on
% set(gca,'fontsize',14,'PlotBoxAspectRatio',[1.4 1 1]);
% grid on;
% xlabel('\fontsize{14}\it Iteration'); ylabel('\fontsize{14}\it Normalized residual');
% legend('\fontsize{14}\it C-LMaFit','\fontsize{14}\it D-LMaFit');
% hold on
% % saveas(gcf,'Pic_200_2000_9','fig')
