clear all; close all; clc;


N = 5; % Number of nodes
noiseFac = 1e-6; % Noise factor
CN = 0; % 0 for well-conditioned


m = 200;
n = 1000;
truerank = 5;
OS = 8;
r = truerank;
 

%% Data generation

M = OS*truerank*(m + n -truerank); % total entries

% The left and right factors which make up our true data matrix Y.
YL = randn(m, truerank);
YR = randn(n, truerank);

% Condition number
if CN > 0
    YLQ = orth(YL);
    YRQ = orth(YR);
        
    s1 = 1000;
    %     step = 1000; S0 = diag([s1:step:s1+(truerank-1)*step]*1); % Linear decay
    S0 = s1*diag(logspace(-log10(CN),0,truerank)); % Exponential decay
    
    YL = YLQ*S0;
    YR = YRQ;
    
    fprintf('Creating a matrix with singular values...\n')
    for kk = 1: length(diag(S0));
        fprintf('%s \n', num2str(S0(kk, kk), '%10.5e') );
    end
    
end



% Select a random set of M entries of Y.
idx = unique(ceil(m*n*rand(1,(10*M))));

idx = idx(randperm(length(idx)));

[I, J] = ind2sub([m, n],idx(1:M));
[J, inxs] = sort(J); I=I(inxs)';

% Values of Y at the locations indexed by I and J.
S = sum(YL(I,:).*YR(J,:), 2);
S_noiseFree = S;

% Add noise.
noise = noiseFac*max(S)*randn(size(S));
S = S + noise;


PA = sparse(I,J,S,m ,n);
P = sparse(I,J,1,m,n);




%% Decentralize LMaFit matrix completion
options.iters = 100; 
[~, ~, infos_dlmafit] = dlmafit_mod(m, n, r, N, P, YL*YR', options);








%%  Plots

pp = 10;

fs = 20;
figure;
semilogy(infos_dlmafit{1}.cost(1:pp:end),'-O','Color','r','LineWidth',2, 'MarkerSize',13);
hold on;
semilogy(infos_dlmafit{2}.cost(1:pp:end),'-s','Color','r','LineWidth',2, 'MarkerSize',13);
semilogy(infos_dlmafit{1}.commcost(1:pp:end),'-+','Color','r','LineWidth',2, 'MarkerSize',13);
hold off;
ax1 = gca;
set(ax1,'FontSize',fs);
xlabel(ax1,['Every ',num2str(pp),'th update'],'FontSize',fs);
ylabel(ax1,'Mean square error on training set','FontSize',fs);
legend('Node 1 cost', 'Node 2 cost','Nodes 1 - 2 comm. cost');
legend 'boxoff';
title([num2str(m),' by ',num2str(n),', rank ',num2str(r),', OS ',num2str(OS)])
box off;






