function W = mylsqfit(U, ii, problem, options)
%
% Adapted from the RTRMC code. It computes the least-squares solution
% for W given U.
%

% Part of the codes for Riemannian gossip algorithms for 
% decentralized matrix completion.
% Original author: Nicolas Boumal <nicolasboumal@gmail.com>. 
% Contributors:    Bamdev Mishra <bamdevm@gmail.com> and
%                  Hiroyuki Kasai <kasai@is.uec.ac.jp>.
 
    % New approach.
    entries = problem.entries_cells{ii};
    masktemp = problem.masks_cells{ii};
    setsparseentries(masktemp, sqrt((1 - (options.lambda).^2)*ones(size(entries)))); % BM: modifies for regaularization
    Achol = spbuildmatrix(masktemp, U, (options.lambda).^2); % BM: modifies for regularization.
    rhs = multfullsparse(U.', entries, masktemp);
    W = (blocksolve(Achol, rhs))';
    
       
    %     % Debug: old approach.
    %     values = problem.values_cells{ii};
    %     indicator = problem.indicator_cells{ii};
    %     W1 = zeros(size(values, 2), size(U, 2));
    %     for k = 1 : size(values, 2),
    %         % Pull out the relevant indices and revealed entries for this column
    %         idx = find(indicator(:,k));
    %         values_Omega = values(idx,k);
    %         U_Omega = U(idx,:);
    %
    %         % Solve a simple least squares problem to populate U
    %         W1(k,:) = (U_Omega\values_Omega)';
    %     end
    %     norm(W1 - W, 'fro')
end



function X = blocksolve(R, B)
    % R is a cell containing n upper triangular square matrices of size r.
    % B is a matrix of size r x n. X is a matrix of size r x n where the i-th
    % column is the solution of the equation A{i}x = B(:, i), with
    % A{i} = R{i}.'*R{i}. This is: the R{i} matrices are the upper triangular
    % Cholesky factors of the matrices A{i} (which we do not require/store).
    
    X = cholsolvecell(R, B);
    
end