clear; clc; close all;

%% Load MovieLens 20M data that has users as rows and movies as columns.
load('/Users/bamdevm/Dropbox/datasets/ml_20m_extracted/data_ml_20m.mat');
data_real_org = data_ml_20m;
data_real.rows = data_real_org.cols; % Transpose operation % Movies

data_real.cols = data_real_org.rows; % Users as columns.
data_real.entries = data_real_org.entries; % Ratings

% Movies as rows. Replace movie ids by integers.
[rows_sorted, IA, IC] = unique(data_real.rows);
movie_ids = 1:length(rows_sorted);
tmp = movie_ids(IC);
data_real.rows = tmp;




%% Remove the ones with entries zero
rm_ind = find(data_real.entries == 0);
data_real.entries(rm_ind) =[];
data_real.rows(rm_ind) =[];
data_real.cols(rm_ind) =[];




%% permute them
random_order = randperm(length(data_real.rows));
data_real.rows = data_real.rows(random_order);
data_real.cols = data_real.cols(random_order);
data_real.entries = data_real.entries(random_order);



%%
m = max(data_real.rows);
n = max(data_real.cols);
r = 3;
rho = 1e7;

N = 4; % Number of nodes
iters = 800; 



%% Train / test split
prop_known = 0.8; % or 0.9 % Training fraction

% Data training
allidx = 1 : length(data_real.rows); % all indices

M = round(prop_known*length(data_real.rows)); % Number of entries known
idx = 1 : M; % indices for training
Matcom.row = (data_real.rows(idx))';
Matcom.col = (data_real.cols(idx))';
Matcom.values_Omega = (data_real.entries(idx))';

S = sparse(Matcom.row', Matcom.col', Matcom.values_Omega', m, n);
[I_0, J_0, data_0] = find(S);
Matcom.row = I_0';
Matcom.col = J_0';
Matcom.values_Omega = data_0';
M = length(Matcom.row);


data_ls.rows = Matcom.row';
data_ls.cols = Matcom.col';
data_ls.entries = Matcom.values_Omega';
data_ls.nentries = length(data_ls.rows);


% Test data
M_test = length(data_real.rows) - M;
L = length(data_real.rows);
idx_test = allidx(L - M_test + 1: L);

data_ts.rows = (data_real.rows(idx_test));
data_ts.cols = (data_real.cols(idx_test));
data_ts.entries = (data_real.entries(idx_test));

%
PA = sparse(data_ls.rows, data_ls.cols, data_ls.entries, m ,n);
P = sparse(data_ls.rows, data_ls.cols, 1, m ,n);

PA_test = sparse(data_ts.rows, data_ts.cols, data_ts.entries, m ,n);
P_test = sparse(data_ts.rows, data_ts.cols, 1, m ,n);


%% Nodes specific data



% ni = [5133 10000 10000 10000 10000 10000 10000]; N = length(ni);
% for ii = 1 : N
%     values_cells{ii} = mat2cell(PA, [m], [ni]);
%     indicator_cells{ii} = mat2cell(P, [m], [ni]);
%     
%     values_cells_test{ii} = mat2cell(PA_test, [m], [ni]);
%     indicator_cells_test{ii} = mat2cell(P_test, m, ni);
% end

l = ceil(n/N);
for ii = 1 : N
    values_cells{ii} = PA(:, 1 + (ii-1)*l: min(ii*l, n));
    indicator_cells{ii} = P(:, 1 + (ii-1)*l: min(ii*l, n));
    
    values_cells_test{ii} = PA_test(:, 1 + (ii-1)*l: min(ii*l, n));
    indicator_cells_test{ii} = P_test(:, 1 + (ii-1)*l: min(ii*l, n));
end




%%

% Initialization of U at all the N agents
problem.M = grassmannfactory(m, r); %  Define the Grassmann manifold
for ii = 1 : N
    U_cells0{ii} = problem.M.rand();
end



% problem structure definition
problem.m = m;
problem.r = r;
problem.N = N;

% train data
for ii = 1 : N
    values = values_cells{ii};
    indicator = indicator_cells{ii};
    [I, J] = find(indicator == 1);
    values_vec = values(:);
    
    problem.I_cells{ii} = uint32(I);
    problem.J_cells{ii} = uint32(J);
    problem.entries_cells{ii} = values_vec(sub2ind(size(values), I, J));
    problem.nentries_cells{ii} = length(problem.entries_cells{ii});
    problem.masks_cells{ii} = sparse(double(I), double(J), ...
        ones(length(I), 1), size(values, 1), size(values, 2), length(I));
end



% test data
for ii = 1:N
    values_test = values_cells_test{ii};
    indicator_test = indicator_cells_test{ii};
    [I_test, J_test] = find(indicator_test == 1);
    values_vec_test = values_test(:);
    
    problem.I_test_cells{ii} = uint32(I_test);
    problem.J_test_cells{ii} =  uint32(J_test);
    problem.entries_test_cells{ii} = values_vec_test(sub2ind(size(values_test), I_test, J_test));
    problem.nentries_test_cells{ii} = length(problem.entries_test_cells{ii});
    
end


%% Call decentralized Grassmann
clear options

options.update = @Grassmann_update;
options.rho = rho;
options.iters = iters;
options.computeNMAE = true;
% options.stepsize = 10^3;


fprintf('\n---------------------  Decentralized Grassmann ---------------------\n')
[U_cells, infos] = online_decentralized_MC(U_cells0, problem, options);



%% Global
global_nmae = 0;
global_test = 0;
for ii = 1:N
   global_nmae = global_nmae + problem.nentries_test_cells{ii}*infos{ii}.nmae_test(end);
   global_test = global_test + problem.nentries_test_cells{ii}*infos{ii}.cost_test(end);
end
global_nmae = global_nmae/length(data_ts.entries)
global_test = global_test/length(data_ts.entries)




%% Plots
pp = 10;

fs = 20;
figure;
semilogy(infos{1}.cost(1:pp:end),'-O','Color','r','LineWidth',2, 'MarkerSize',13);
hold on;
semilogy(infos{2}.cost(1:pp:end),'-s','Color','r','LineWidth',2, 'MarkerSize',13);
semilogy(infos{1}.commcost(1:pp:end),'-+','Color','m','LineWidth',2, 'MarkerSize',13);
semilogy(infos{2}.commcost(1:pp:end),'-*','Color','b','LineWidth',2, 'MarkerSize',13);
semilogy(infos{3}.commcost(1:pp:end),'-d','Color','k','LineWidth',2, 'MarkerSize',13);
hold off;
ax1 = gca;
set(ax1,'FontSize',fs);
xlabel(ax1,['Every ',num2str(pp),'th update'],'FontSize',fs);
ylabel(ax1,'Mean square error on training set','FontSize',fs);
legend('Agent 1 cost', 'Agent 2 cost','Nodes 1 - 2 comm. cost', 'Nodes 2 - 3 comm. cost','Nodes 3 - 4 comm. cost');
legend 'boxoff';
title([num2str(m),' by ',num2str(n),', rank ',num2str(r)])
box off;




fs = 20;
figure;
plot(infos{1}.cost_test(1:pp:end),'-O','Color','r','LineWidth',2, 'MarkerSize',13);
hold on;
plot(infos{2}.cost_test(1:pp:end),'-s','Color','r','LineWidth',2, 'MarkerSize',13);
hold off;
ax1 = gca;
set(ax1,'FontSize',fs);
xlabel(ax1,['Every ',num2str(pp),'th update'],'FontSize',fs);
ylabel(ax1,'Mean square error on test set','FontSize',fs);
legend('Agent 1 cost', 'Agent 2 cost');
legend 'boxoff';
title([num2str(m),' by ',num2str(n),', rank ',num2str(r)])
box off;





