clear all; close all; clc;

iters = 1000;
N = 6; % Number of agents
rho1 = 1e3; 
rho2 = 1e10;

noiseFac = 1e-6; % Noise factor
CN = 0; % 0 for well-conditioned

m = 10000;
n = 100000;
truerank = 5;
OS = 6;
r = truerank;


%% Data generation

M = OS*truerank*(m + n -truerank); % total entries

% The left and right factors which make up our true data matrix Y.
YL = randn(m, truerank);
YR = randn(n, truerank);

% Condition number
if CN > 0
    YLQ = orth(YL);
    YRQ = orth(YR);
    
    s1 = 1000;
    %     step = 1000; S0 = diag([s1:step:s1+(truerank-1)*step]*1); % Linear decay
    S0 = s1*diag(logspace(-log10(CN),0,truerank)); % Exponential decay
    
    YL = YLQ*S0;
    YR = YRQ;
    
    fprintf('Creating a matrix with singular values...\n')
    for kk = 1: length(diag(S0));
        fprintf('%s \n', num2str(S0(kk, kk), '%10.5e') );
    end
    
end



% Select a random set of M entries of Y.
idx = unique(ceil(m*n*rand(1,(10*M))));
idx = idx(randperm(length(idx)));

[I, J] = ind2sub([m, n],idx(1:M));
[J, inxs] = sort(J); I=I(inxs)';

% Values of Y at the locations indexed by I and J.
S = sum(YL(I,:).*YR(J,:), 2);
S_noiseFree = S;

% Add noise.
noise = noiseFac*max(S)*randn(size(S));
S = S + noise;


PA = sparse(I,J,S,m ,n);
P = sparse(I,J,1,m,n);




%%

% Nodes specific data
l = ceil(n/N);
for ii = 1 : N
    values_cells{ii} = PA(:, 1 + (ii-1)*l: min(ii*l, n));
    indicator_cells{ii} = P(:, 1 + (ii-1)*l: min(ii*l, n));
end


% Initialization of U at all the N agents
problem.M = grassmannfactory(m, r); %  Define the Grassmann manifold
for ii = 1 : N
    U_cells0{ii} = problem.M.rand();
end



%% problem structure definition
problem.m = m;
problem.r = r;
problem.N = N;

for ii = 1 : N
    values = values_cells{ii};
    indicator = indicator_cells{ii};
    [I, J] = find(indicator == 1);
    values_vec = values(:);
    
    problem.I_cells{ii} = uint32(I);
    problem.J_cells{ii} = uint32(J);
    problem.entries_cells{ii} = values_vec(sub2ind(size(values), I, J));
    problem.nentries_cells{ii} = length(problem.entries_cells{ii});
    problem.masks_cells{ii} = sparse(double(I), double(J), ...
        ones(length(I), 1), size(values, 1), size(values, 2), length(I));
end



%% test data
for ii = 1:N
    [m_ii, n_ii] = size(values_cells{ii});
    test_nentries = 2*length(problem.I_cells{ii});
    
    I_test = uint32(randi(m_ii, test_nentries, 1));
    J_test = uint32(randi(n_ii, test_nentries, 1));
    
    YR_ii = YR(1 + (ii-1)*l: min(ii*l, n), :);
    
    entries_test = spmaskmult(YL, YR_ii', uint32(I_test), uint32(J_test));
    
    problem.I_test_cells{ii} = I_test;
    problem.J_test_cells{ii} = J_test;
    problem.entries_test_cells{ii} = entries_test;
    problem.nentries_test_cells{ii} = length(problem.entries_test_cells{ii});
    
end


%% Decentralized Grassmann
options.rho = rho1;
options.update = @Grassmann_update;
options.iters = iters;
options.stepsize = 1e-3;

fprintf('\n---------------------  Decentralized Grassmann ---------------------\n')
[U_cells, infos] = online_decentralized_MC(U_cells0, problem, options);

clear options;
 









%% Decentralized Grassmann
options.rho = rho2;
options.update = @Grassmann_update;
options.iters = iters;

fprintf('\n---------------------  Decentralized Grassmann ---------------------\n')
[U_cells1, infos1] = online_decentralized_MC(U_cells0, problem, options);

clear options;











%% Plots

pp = 10;

fs = 20;
figure;
semilogy(infos{1}.cost(1:pp:end),'-O','Color','r','LineWidth',2, 'MarkerSize',13);
hold on;
semilogy(infos{2}.cost(1:pp:end),'-s','Color','r','LineWidth',2, 'MarkerSize',13);
semilogy(infos{1}.commcost(1:pp:end),'-+','Color','r','LineWidth',2, 'MarkerSize',13);
semilogy(infos1{1}.cost(1:pp:end),':O','Color','b','LineWidth',2, 'MarkerSize',13);
semilogy(infos1{2}.cost(1:pp:end),':s','Color','b','LineWidth',2, 'MarkerSize',13);
semilogy(infos1{1}.commcost(1:pp:end),':+','Color','b','LineWidth',2, 'MarkerSize',13);
hold off;
ax1 = gca;
set(ax1,'FontSize',fs);
xlabel(ax1,['Every ',num2str(pp),'th update'],'FontSize',fs);
ylabel(ax1,'Mean square error on training set','FontSize',fs);
legend(['Agent 1 rho=10^{3}'], ['Agent 2, rho=10^{3}'], ['Distance 1-2, rho=10^{3}'],['Agent 1, rho=10^{10}'], ['Agent 2, rho=10^{10} '], ['Distance 1-2, rho=10^{10}']);
legend 'boxoff';
title([num2str(m),' by ',num2str(n),', rank ',num2str(r),', OS ',num2str(OS)])
box off;






